module Main where

import Test.QuickCheck

data Exp = ANum Int
         | APlus Exp Exp
         | ASub  Exp Exp
         | AMult Exp Exp
         deriving (Eq, Show)

eval :: Exp -> Int
eval (ANum n) = n
eval (APlus e1 e2) = eval e1 + eval e2
eval (ASub  e1 e2) = eval e1 - eval e2
eval (AMult e1 e2) = eval e1 * eval e2

optimize :: Exp -> Exp
optimize (ANum n) = ANum n
optimize (APlus e (ANum 0)) = e
optimize (APlus (ANum 0) e) = e
optimize (ASub e (ANum 0)) = e
--optimize (ASub (ANum 0) e) = e
optimize (AMult e (ANum 1)) = e
optimize (AMult (ANum 1) e) = e
optimize e = e

prop_OptimizeCorrect :: Exp -> Bool
prop_OptimizeCorrect e = eval (optimize e) == eval e

genExp :: Int -> Gen Exp
genExp 0 = ANum <$> arbitrary
genExp n = oneof [ ANum  <$> arbitrary
                 , APlus <$> genExp (n `div` 2) <*> genExp (n `div` 2)
                 , ASub  <$> genExp (n `div` 2) <*> genExp (n `div` 2)
                 , AMult <$> genExp (n `div` 2) <*> genExp (n `div` 2)
                 ]

shrinkExp :: Exp -> [Exp]
shrinkExp (ANum 0) = []
shrinkExp (ANum n) = ANum <$> shrink n
shrinkExp (APlus e1 e2) = e1 : e2
                        :  [APlus e1' e2 | e1' <- shrink e1]
                        ++ [APlus e1 e2' | e2' <- shrink e2]
shrinkExp (ASub e1 e2) = e1 : e2
                        :  [ASub e1' e2 | e1' <- shrink e1]
                        ++ [ASub e1 e2' | e2' <- shrink e2]
shrinkExp (AMult e1 e2) = e1 : e2
                        :  [AMult e1' e2 | e1' <- shrink e1]
                        ++ [AMult e1 e2' | e2' <- shrink e2]


  


instance Arbitrary Exp where
  arbitrary = sized genExp
  shrink = shrinkExp

main :: IO ()
main = quickCheckWith stdArgs{maxSuccess = 10000} prop_CompileCorrect

-- Stack Machine Instructions

data SInstr = SPush Int | SPlus | SSub | SMult
            deriving (Eq, Show)

exec :: [Int] -> [SInstr] -> [Int]
exec stack [] = stack
exec stack (SPush n:prog) = exec (n : stack) prog
exec (m:n:stack) (SPlus: prog) = exec ((m+n):stack) prog
exec (m:n:stack) (SSub : prog) = exec ((m-n):stack) prog
exec (m:n:stack) (SMult: prog) = exec ((m*n):stack) prog

compile :: Exp -> [SInstr]
compile (ANum n) = [SPush n]
compile (APlus e1 e2) = compile e2 ++ compile e1 ++ [SPlus]
compile (ASub  e1 e2) = compile e2 ++ compile e1 ++ [SSub]
compile (AMult e1 e2) = compile e2 ++ compile e1 ++ [SMult]


prop_CompileCorrect :: Exp -> Bool
prop_CompileCorrect e = exec [] (compile e) == [eval e]

